#include "Translate.h"

#include <IGESControl_Reader.hxx>
#include <IGESControl_Writer.hxx>
#include <IGESControl_Controller.hxx>
#include <STEPControl_Reader.hxx>
#include <STEPControl_Writer.hxx>
#include <STEPControl_StepModelType.hxx>
#include <Interface_Static.hxx>

#include <StlAPI_Writer.hxx>
#include <VrmlAPI_Writer.hxx>

#include <BRepTools.hxx>
#include <BRep_Tool.hxx>
#include <BRep_Builder.hxx>

#include <TopoDS.hxx>
#include <TopoDS_Shape.hxx>
#include <TopoDS_Compound.hxx>
#include <TopExp_Explorer.hxx>
#include <TopTools_HSequenceOfShape.hxx>

#include <Geom_Line.hxx>
#include <Geom_Curve.hxx>
#include <Geom_Plane.hxx>
#include <Geom_Surface.hxx>

using namespace std;

// ---------------------------- Translate -----------------------------------------

Translate::Translate( )
{
}

Translate::~Translate()
{
}

Handle(TopTools_HSequenceOfShape) Translate::importModel( const int format, const string& file )
{
    Handle(TopTools_HSequenceOfShape) shapes;
    
	try 
	{
        switch ( format )
        {
        case FormatBREP:
            shapes = importBREP( file );
            break;
        case FormatIGES:
            shapes = importIGES( file );
            break;
        case FormatSTEP:
            shapes = importSTEP( file );
            break;
        }
    }
	
    catch ( const Standard_Failure & ) 
	{
        shapes.Nullify();
    }
    return shapes;
}

bool Translate::exportModel( const int format, const string& file, const Handle(TopTools_HSequenceOfShape)& shapes )
{
    try 
	{
        switch ( format )
        {
			case FormatBREP: return exportBREP( file, shapes );
			case FormatIGES: return exportIGES( file, shapes );
			case FormatSTEP: return exportSTEP( file, shapes );
			case FormatSTL:  return exportSTL ( file, shapes );
			case FormatVRML: return exportVRML( file, shapes );
        }
    } 
    
	catch ( const Standard_Failure & ) 
	{
        clog << "ERROR_IMPORTING: " << file << endl;
    }
    
	return false;
}

// ----------------------------- Import functionality -----------------------------

Handle(TopTools_HSequenceOfShape) Translate::importBREP( const string& file )
{
	Handle(TopTools_HSequenceOfShape) aSequence;
    TopoDS_Shape aShape;
	BRep_Builder aBuilder;
	TCollection_AsciiString  aFilePath = file.c_str();
	Standard_Boolean result = BRepTools::Read( aShape, aFilePath.ToCString(), aBuilder );
	
	if ( result )
    {
	    aSequence = new TopTools_HSequenceOfShape();
		aSequence->Append( aShape );
    }
	
    return aSequence;
}

Handle(TopTools_HSequenceOfShape) Translate::importIGES( const string& file )
{
    Handle(TopTools_HSequenceOfShape) aSequence;
    TCollection_AsciiString  aFilePath = file.c_str();
    
    IGESControl_Reader Reader;
    int status = Reader.ReadFile(file.c_str());

    if ( status == IFSelect_RetDone )
    {
        aSequence = new TopTools_HSequenceOfShape();
        Reader.TransferRoots();
        TopoDS_Shape aShape = Reader.OneShape();
        aSequence->Append( aShape );
    }
	
	return aSequence;
}

Handle(TopTools_HSequenceOfShape) Translate::importSTEP( const string& file )
{
    Handle(TopTools_HSequenceOfShape) aSequence = new TopTools_HSequenceOfShape;
    TCollection_AsciiString  aFilePath = file.c_str();
    STEPControl_Reader aReader;
    IFSelect_ReturnStatus status = aReader.ReadFile( aFilePath.ToCString() );
    
	if ( status != IFSelect_RetDone )
        return aSequence;

    //Interface_TraceFile::SetDefault();
    bool failsonly = false;
    aReader.PrintCheckLoad( failsonly, IFSelect_ItemsByEntity );

    int nbr = aReader.NbRootsForTransfer();
    aReader.PrintCheckTransfer( failsonly, IFSelect_ItemsByEntity );
    
	for ( Standard_Integer n = 1; n <= nbr; n++ )
        aReader.TransferRoot( n );

    int nbs = aReader.NbShapes();
    
	if ( nbs > 0 )
    {
        for ( int i = 1; i <= nbs; i++ )
        {
			TopoDS_Shape shape = aReader.Shape( i );
			aSequence->Append( shape );
        }
    }

    return aSequence;
}

// ----------------------------- Export functionality -----------------------------

bool Translate::exportBREP( const string& file, const Handle(TopTools_HSequenceOfShape)& shapes )
{
    if ( shapes.IsNull() || shapes->IsEmpty() )
        return false;

    TopoDS_Shape shape = shapes->Value( 1 );
    
    const TCollection_AsciiString anUtf8Path (file.c_str());
    
    return BRepTools::Write( shape, anUtf8Path.ToCString() ); 
}

bool Translate::exportIGES( const string& file, const Handle(TopTools_HSequenceOfShape)& shapes )
{
    if ( shapes.IsNull() || shapes->IsEmpty() )
        return false;

    IGESControl_Controller::Init();
	IGESControl_Writer writer( Interface_Static::CVal( "XSTEP.iges.unit" ),
                               Interface_Static::IVal( "XSTEP.iges.writebrep.mode" ) );
 
	for ( int i = 1; i <= shapes->Length(); i++ )
		writer.AddShape ( shapes->Value( i ) );
	writer.ComputeModel();
	
	const TCollection_AsciiString anUtf8Path (file.c_str());
	return writer.Write( anUtf8Path.ToCString());
}

bool Translate::exportSTEP( const string& file, const Handle(TopTools_HSequenceOfShape)& shapes )
{
    if ( shapes.IsNull() || shapes->IsEmpty() )
        return false;

    STEPControl_StepModelType type = STEPControl_AsIs; 
	// also possible types: STEPControl_ManifoldSolidBrep, STEPControl_BrepWithVoids, STEPControl_FacetedBrep, TEPControl_FacetedBrepAndBrepWithVoids, 
	// STEPControl_ShellBasedSurfaceModel, STEPControl_GeometricCurveSet, STEPControl_Hybrid (see OCCT documentation)
    
	if ( type < 0 )
        return false;
    
    IFSelect_ReturnStatus status;

    if ( type == STEPControl_FacetedBrep && !checkFacetedBrep( shapes ) )
    {
        clog << "INF_FACET_ERROR" << endl;
        return false;
    }

    STEPControl_Writer writer;
	
	for ( int i = 1; i <= shapes->Length(); i++ )
    {
		status = writer.Transfer( shapes->Value( i ), type );
        if ( status != IFSelect_RetDone )
            return false;
    }
    
    const TCollection_AsciiString anUtf8Path (file.c_str());
    status = writer.Write( anUtf8Path.ToCString() );

    switch ( status )
    {
    case IFSelect_RetError:
        clog << "INF_DATA_ERROR" << endl;
        break;
    case IFSelect_RetFail:
        clog << "INF_WRITING_ERROR" << endl;
        break;
    case IFSelect_RetVoid:
        clog << "INF_NOTHING_ERROR" << endl;
        break;
    case IFSelect_RetStop:
    case IFSelect_RetDone:
        break;
    }
    return status == IFSelect_RetDone;
}

bool Translate::exportSTL( const string& file, const Handle(TopTools_HSequenceOfShape)& shapes )
{
    if(shapes.IsNull() || shapes->IsEmpty()) return false;
    
	TopoDS_Compound res;
	BRep_Builder builder;
	builder.MakeCompound( res );

	for ( int i = 1; i <= shapes->Length(); i++ )
	{
		TopoDS_Shape shape = shapes->Value( i );
		if ( shape.IsNull() ) 
		{
			clog << "INF_TRANSLATE_ERROR_INVALIDSHAPE" << endl;
			return false;
        }
		builder.Add( res, shape );
	}

	StlAPI_Writer writer;
	const TCollection_AsciiString anUtf8Path (file.c_str());
	writer.Write( res, anUtf8Path.ToCString());

    return true;
}

bool Translate::exportVRML( const string& file, const Handle(TopTools_HSequenceOfShape)& shapes )
{
    if ( shapes.IsNull() || shapes->IsEmpty() ) return false;
    
	TopoDS_Compound res;
	BRep_Builder builder;
	builder.MakeCompound( res );

	for ( int i = 1; i <= shapes->Length(); i++ )
	{
		TopoDS_Shape shape = shapes->Value( i );
		if ( shape.IsNull() )
		{
			clog << "INF_TRANSLATE_ERROR_INVALIDSHAPE" << endl;
			return false;
        }
		builder.Add( res, shape );
	}

	VrmlAPI_Writer writer;
	const TCollection_AsciiString anUtf8Path (file.c_str());
	writer.Write( res, anUtf8Path.ToCString() );

    return true;
}

bool Translate::checkFacetedBrep( const Handle(TopTools_HSequenceOfShape)& shapes )
{
	bool err = false;
	
	for(int i=1; i<=shapes->Length(); i++)
	{
	    TopoDS_Shape shape = shapes->Value(i);
		
        for ( TopExp_Explorer fexp( shape, TopAbs_FACE ); fexp.More() && !err; fexp.Next() )
		{
		    Handle(Geom_Surface) surface = BRep_Tool::Surface( TopoDS::Face( fexp.Current() ) );
		    if ( !surface->IsKind( STANDARD_TYPE( Geom_Plane ) ) ) err = true;
		}
		
        for ( TopExp_Explorer eexp( shape, TopAbs_EDGE ); eexp.More() && !err; eexp.Next() )
		{
		    Standard_Real fd, ld;
		    Handle(Geom_Curve) curve = BRep_Tool::Curve( TopoDS::Edge( eexp.Current() ), fd, ld );
		    if ( !curve->IsKind( STANDARD_TYPE( Geom_Line ) ) ) err = true;
		}
	}
	
	return !err;
}



